@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package Test::Mojo::Role::OpenAPI::Modern; # git description: v0.003-3-ga41ead5
+package Test::Mojo::Role::OpenAPI::Modern; # git description: v0.004-7-g1a03221
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Test::Mojo role providing access to an OpenAPI document and parser
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.004';
+our $VERSION = '0.005';
 
 use 5.020;  # for fc, unicode_strings features
 use strictures 2;
@@ -13,6 +13,9 @@ use if "$]" >= 5.022, experimental => 're_strict';
 no if "$]" >= 5.031009, feature => 'indirect';
 no if "$]" >= 5.033001, feature => 'multidimensional';
 no if "$]" >= 5.033006, feature => 'bareword_filehandles';
+use JSON::Schema::Modern 0.577;
+use OpenAPI::Modern 0.054;
+use List::Util 'any';
 use namespace::clean;
 
 use Mojo::Base -role, -signatures;
@@ -39,6 +42,10 @@ has 'openapi' => sub ($self) {
   die 'openapi object or configs required';
 };
 
+# keys being tracked here:
+# - request_result
+# - response_result
+# - validation_options
 sub _openapi_stash ($self, @data) { Mojo::Util::_stash(_openapi_stash => $self, @data) }
 
 after _request_ok => sub ($self, @args) {
@@ -53,12 +60,27 @@ sub response_valid ($self, $desc = 'response is valid') {
   return $self->test('ok', $self->response_validation_result, $desc);
 }
 
-sub request_not_valid ($self, $desc = 'request is not valid') {
-  return $self->test('ok', !$self->request_validation_result, $desc);
+# expected_error can either be the 'recommended_response' or any error string in the result.
+sub request_not_valid ($self, $expected_error = undef, $desc = undef) {
+  my $result = $self->request_validation_result;
+  if ($expected_error and not $result) {
+    $desc //= 'request validation error matches';
+    return $self->test('pass', $desc) if $expected_error eq $result->recommended_response->[1];
+    return $self->test('ok', (any { $expected_error eq $_ } $result->errors), $desc);
+  }
+
+  return $self->test('ok', !$self->request_validation_result, $desc // 'request is not valid');
 }
 
-sub response_not_valid ($self, $desc = 'response is not valid') {
-  return $self->test('ok', !$self->response_validation_result, $desc);
+# expected_error must match an error string in the result.
+sub response_not_valid ($self, $expected_error = undef, $desc = undef) {
+  my $result = $self->response_validation_result;
+  if ($expected_error and not $result) {
+    return $self->test('ok', (any { $expected_error eq $_ } $result->errors),
+      $desc // 'response validation error matches');
+  }
+
+  return $self->test('ok', !$self->response_validation_result, $desc // 'response is not valid');
 }
 
 # I'm not sure yet which names are best, why not go for both?
@@ -67,24 +89,34 @@ sub response_invalid { goto \&response_not_valid }
 
 sub request_validation_result ($self) {
   my $result = $self->_openapi_stash('request_result');
-  return $result if $result;
+  return $result if defined $result;
 
-  $result = $self->openapi->validate_request($self->tx->req);
-  $self->_openapi_stash(request_result => $result);
+  $result = $self->openapi->validate_request($self->tx->req, my $options = {});
+  $self->_openapi_stash(request_result => $result, validation_options => $options);
   return $result;
 }
 
 sub response_validation_result ($self) {
   my $result = $self->_openapi_stash('response_result');
-  return $result if $result;
+  return $result if defined $result;
+
+  my $options = $self->_openapi_stash('validation_options');
+  $self->_openapi_stash(validation_options => $options = {}) if not $options;
+  $options->{request} = $self->tx->req;
 
-  my $options = { request => $self->tx->req };
   $result = $self->openapi->validate_response($self->tx->res, $options);
-
   $self->_openapi_stash(response_result => $result);
   return $result;
 }
 
+sub operation_id_is ($self, $operation_id, $desc = undef) {
+  my $validation_options = $self->_openapi_stash('validation_options');
+  return $self->test('fail', 'operation_id was captured during the last validation')
+    if not $validation_options->{operation_id};
+
+  $self->test('is', $validation_options->{operation_id}, $operation_id, $desc // 'operation_id is '.$operation_id);
+}
+
 1;
 
 __END__
@@ -99,7 +131,7 @@ Test::Mojo::Role::OpenAPI::Modern - Test::Mojo role providing access to an OpenA
 
 =head1 VERSION
 
-version 0.004
+version 0.005
 
 =head1 SYNOPSIS
 
@@ -118,7 +150,7 @@ version 0.004
         in: path
         required: true
         schema:
-          pattern: ^[a-z]+$
+          type: integer
       post:
         operationId: my_foo_request
         requestBody:
@@ -144,11 +176,21 @@ version 0.004
     ->new('MyApp', { ... })
     ->openapi($openapi);
 
-  $t->post_ok('/foo/hello')
+  $t->post_ok('/foo/123')
     ->status_is(200)
     ->json_is('/status', 'ok')
     ->request_valid
-    ->response_valid;
+    ->response_valid
+    ->operation_id_is('my_foo_request');
+
+  $t->post_ok('/foo/123', form => { salutation => 'hi' })
+    ->status_is(400)
+    ->request_not_valid('Unsupported Media Type')
+    ->response_not_valid(q{'/response': no response object found for code 400});
+
+  $t->post('/foo/hello')
+    ->status_is(400)
+    ->request_not_valid(q{/request/path: 'wrong type: expected integer, got string'}, 'detected bad path parameter');
 
 =head1 DESCRIPTION
 
@@ -177,7 +219,8 @@ L<Mojolicious::Plugin::OpenAPI::Modern/validate_request>, producing a boolean te
 
 =head2 request_not_valid, request_invalid
 
-The inverse of L</request_valid>.
+The inverse of L</request_valid>. May optionally take an error string, which is matched against
+the L<JSON::Schema::Modern::Result/recommended_response>, and all error strings in the result.
 
 =head2 response_valid
 
@@ -186,7 +229,18 @@ L<Mojolicious::Plugin::OpenAPI::Modern/validate_response>, producing a boolean t
 
 =head2 response_not_valid, response_invalid
 
-The inverse of L</response_valid>.
+The inverse of L</response_valid>. May optionally take an error string, which is matched against all
+error strings in the result.
+
+Note that normally you shouldn't be testing for an invalid response, as all responses from your
+application should be valid according to the specification, but this method is provided for
+completeness. Instead, check for invalid responses right in your application (see
+L<Mojolicious::Plugin::OpenAPI::Modern>) and log an error when this occurs.
+
+=head2 operation_id_is
+
+Test the C<operationId> corresponding to the last validated request or response (the unique string
+used to identify the operation). See L<https://spec.openapis.org/oas/v3.1.0#operation-object>.
 
 =head2 request_validation_result
 
